# Copyright (c) Microsoft Corporation
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY
# IMPLIED WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR
# PURPOSE, MERCHANTABLITY OR NON-INFRINGEMENT.
#
# See the Apache Version 2.0 License for specific language governing
# permissions and limitations under the License.

"""
Exception classes representing error responses from the Yammer API.
"""

class ResponseError(Exception):
    """
    Raised when the Yammer API responds with an HTTP error, and there
    isn't a more specific subclass that represents the situation.
    """
    pass


class NotFoundError(ResponseError):
    """
    Raised when the Yammer API responds with an HTTP 404 Not Found error.
    """
    pass


class UnauthorizedError(ResponseError):
    """
    Raised when the Yammer API responds with an HTTP 401 Unauthorized error.
    This may mean that the access token you are using is invalid.
    """
    pass


class InvalidAccessTokenError(ResponseError):
    """
    Raised when a request is made with an access token that has expired or has
    been revoked by the user.
    """
    pass


class RateLimitExceededError(ResponseError):
    """
    Raised when a request is rejected because the rate limit has been
    exceeded.
    """
    pass


class InvalidMessageError(Exception):
    """
    Super class for the various kinds of errors that can occur when creating
    a message.
    """
    pass


class TooManyTopicsError(InvalidMessageError):
    """
    Raised when a message cannot be created because too many topics have been
    specified.
    """
    pass


class InvalidOpenGraphObjectError(InvalidMessageError):
    """
    Raised when an invalid Open Graph object is attached to a new message.
    """
    pass

class InvalidUserError(Exception):
    """
    Super class for the various kinds of errors that can occur when creating
    a user.
    """
    pass

class InvalidEducationRecordError(InvalidUserError):
    """
    Raised when creating a user with an education record that doesn't include
    all of the requried information.
    """
    pass

class InvalidPreviousCompanyRecord(InvalidUserError):
    """
    Raised when creating a user with a previous_companies record that doesn't
    include all of the required fields.
    """
    pass



from yampy.apis.utils import ArgumentConverter, none_filter, stringify_booleans
from yampy.models import extract_id


class GroupsAPI(object):

    """
    Provides an interface for accessing the groups related endpoints of the
    Yammer API. You should not instantiate this class directly; use the
    :meth:`yampy.Yammer.groups` method instead.
    """

    def __init__(self, client):
        """
        Initializes a new GroupsAPI that will use the given client object
        to make HTTP requests.
        """
        self._client = client
        self._argument_converter = ArgumentConverter(
            none_filter, stringify_booleans,
        )

    def all(self, mine=None, reverse=None):
        """
        Returns all the groups in the current user's network.

        Customize the response using the keyword arguments:

        * mine -- Only return group of current user.
        * reverse -- return group in descending order by name.
        """
        return self._client.get("/groups", **self._argument_converter(
            mine=mine,
            reverse=reverse,
        ))

    def find(self, group_id):
        """
        Returns the group identified by the given group_id.
        """
        return self._client.get(self._group_path(group_id))

    def members(self, group_id, page=None, reverse=None):
        """
        Returns the group identified by the given group_id.

        Customize the response using the keyword arguments:

        * page -- Enable pagination, and return the nth page of 50 users.
        """
        path = "/group_memberships"
        return self._client.get(path, **self._argument_converter(
            page=page,
            reverse=reverse,
        ))

    def join(self, group_id):
        """
        Join the group identified by the given group_id.

        Return True
        """
        path = "/group_memberships"
        group_id = extract_id(group_id)
        return self._client.post(path, **self._argument_converter(
            group_id=group_id,
        ))

    def leave(self, group_id):
        """
        Leave the group identified by the given group_id.

        Return True
        """
        path = "/group_memberships"
        group_id = extract_id(group_id)
        return self._client.delete(path, **self._argument_converter(
            group_id=group_id,
        ))

    def create(self, name, private=False):
        """
        Create a group.

        Return Group info
        """
        path = "/groups"
        return self._client.post(path, **self._argument_converter(
            name=name,
            private=private,
        ))

    def delete(self, group_id):
        """
        Delete a group.

        Return True if success
        """
        return self._client.delete(self._group_path(group_id), delete="true")

    def _group_path(self, group_id):
        return "/groups/%d" % extract_id(group_id)



# Copyright (c) Microsoft Corporation
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY
# IMPLIED WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR
# PURPOSE, MERCHANTABLITY OR NON-INFRINGEMENT.
#
# See the Apache Version 2.0 License for specific language governing
# permissions and limitations under the License.

import requests

from .constants import DEFAULT_BASE_URL
from .errors import ResponseError, NotFoundError, InvalidAccessTokenError, \
    RateLimitExceededError, UnauthorizedError
from .models import GenericModel


class Client(object):
    """
    A client for the Yammer API.
    """

    def __init__(self, access_token=None, base_url=None):
        self._access_token = access_token
        self._base_url = base_url or DEFAULT_BASE_URL

    def get(self, path, **kwargs):
        """
        Makes an HTTP GET request to the Yammer API. Any keyword arguments will
        be converted to query string parameters.

        The path should be the path of an API endpoint, e.g. "/messages"
        """
        return self._request("get", path, **kwargs)

    def post(self, path, **kwargs):
        """
        Makes an HTTP POST request to the Yammer API. Any keyword arguments
        will be sent as the body of the request.

        The path should be the path of an API endpoint, e.g. "/messages"
        """
        return self._request("post", path, **kwargs)

    def put(self, path, **kwargs):
        """
        Makes an HTTP PUT request to the Yammer API. Any keyword arguments will
        be sent as the body of the request.

        The path should be the path of an API endpoint, e.g. "/users/123"
        """
        return self._request("put", path, **kwargs)

    def delete(self, path, **kwargs):
        """
        Makes an HTTP DELETE request to the Yammer API.

        The path should be the path of an API endpoint, e.g. "/messages/123"
        """
        return self._request("delete", path, **kwargs)

    def _request(self, method, path, **kwargs):
        response = requests.request(
            method=method,
            url=self._build_url(path),
            headers=self._build_headers(),
            params=kwargs,
        )
        return self._parse_response(response)

    def _build_url(self, path):
        return self._base_url + path + ".json"

    def _build_headers(self):
        if self._access_token:
            return {
                "Authorization": "Bearer %s" % self._access_token,
            }
        else:
            return {}

    def _parse_response(self, response):
        if 200 <= response.status_code < 300:
            return self._value_for_response(response)
        else:
            raise self._exception_for_response(response)

    def _value_for_response(self, response):
        if response.text.strip():
            return GenericModel.from_json(response.text)
        else:
            return True

    def _exception_for_response(self, response):
        if response.status_code == 404:
            return NotFoundError(response.reason)
        elif response.status_code == 400 and "OAuthException" in response.text:
            return InvalidAccessTokenError(response.reason)
        elif response.status_code == 401:
            return UnauthorizedError(response.reason)
        elif response.status_code == 429:
            return RateLimitExceededError(response.reason)
        else:
            return ResponseError("%d error: %s" % (
                response.status_code, response.reason,
            ))



# Copyright (c) Microsoft Corporation
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY
# IMPLIED WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR
# PURPOSE, MERCHANTABLITY OR NON-INFRINGEMENT.
#
# See the Apache Version 2.0 License for specific language governing
# permissions and limitations under the License.

import json


class GenericModel(dict):
    """
    A dict subclass that provides access to its members as if they were
    attributes.

    Note that an attribute that has the same name as one of dict's existing
    method (``keys``, ``items``, etc.) will not be accessible as an attribute.
    """

    @classmethod
    def from_json(cls, json_string):
        """
        Parses the given json_string, returning GenericModel instances instead
        of dicts.
        """
        return json.loads(json_string, object_hook=cls)

    def __getattr__(self, prop):
        """
        Provides access to the members of the dict as attributes.
        """
        if prop in self:
            return self[prop]
        else:
            raise AttributeError


def extract_id(source):
    """
    Attempts to extract an ID from the argument, first by looking for an
    attribute and then by using dictionary access. If both fail, the argument
    is returned.
    """
    try:
        return source.id
    except AttributeError:
        pass

    try:
        return source["id"]
    except (TypeError, KeyError) as err:
        pass

    return source



# Copyright (c) Microsoft Corporation
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY
# IMPLIED WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR
# PURPOSE, MERCHANTABLITY OR NON-INFRINGEMENT.
#
# See the Apache Version 2.0 License for specific language governing
# permissions and limitations under the License.

try:
    from urllib.parse import urlencode  # Python 3
except ImportError:
    from urllib import urlencode        # Python 2

from .constants import DEFAULT_OAUTH_BASE_URL, DEFAULT_OAUTH_DIALOG_URL
from .client import Client
from .errors import ResponseError

class Authenticator(object):
    """
    Responsible for authenticating users against the Yammer API.

    The OAuth2 authentication process involves several steps:

    1. Send the user to the URL returned by ``authorization_url``. They can use
       this page to grant your application access to their account.
    2. Yammer redirects them to the ``redirect_uri`` you provided with a code
       that can be exchanged for an access token.
    3. Exchange the code for an access token using the ``fetch_access_token``
       method.
    """

    def __init__(self, client_id, client_secret,
                 oauth_dialog_url=None, oauth_base_url=None):
        """
        Initializes a new Authenticator. The client_id and client_secret
        identify your application, you acquire them when registering your
        application with Yammer. See http://www.yammer.com/client_applications

        Keyword arguments can be used to modify the URLs generated in this
        class, e.g. to avoid hitting the live API from a client application's
        test suite. Pass None to use the default URLs.

        * ``oauth_dialog_url`` -- The URL the user must visit to authorize the
          application. Used by the authorization_url method.
        * ``oauth_base_url`` -- The base URL for OAuth API requests, e.g. token
          exchange. Used by ``fetch_access_data`` or ``fetch_access_token``.
        """
        self._client_id = client_id
        self._client_secret = client_secret
        self._oauth_dialog_url = oauth_dialog_url or DEFAULT_OAUTH_DIALOG_URL
        self._oauth_base_url = oauth_base_url or DEFAULT_OAUTH_BASE_URL

    def authorization_url(self, redirect_uri):
        """
        Returns the URL the user needs to visit to grant your application
        access to their Yammer account. When they are done they will be
        redirected to the ``redirect_uri`` you provide with a code that can be
        exchanged for an access token.
        """
        query = urlencode({
            "client_id": self._client_id,
            "redirect_uri": redirect_uri,
        })
        return "?".join([self._oauth_dialog_url, query])

    def fetch_access_data(self, code):
        """
        Returns the complete response from the Yammer API access token request.
        This is a dict with "user", "network" and "access_token" keys.

        You can access the token itself as: ``response.access_token.token``

        If you only intend to make use of the token, you can use the
        ``fetch_access_token`` method instead for convenience.
        """
        client = Client(base_url=self._oauth_base_url)
        return client.get(
            path="/access_token",
            client_id=self._client_id,
            client_secret=self._client_secret,
            code=code,
        )

    def fetch_access_token(self, code):
        """
        Convenience method to exchange a code for an access token, discarding
        the other user and network data that the Yammer API returns with the
        access token.

        If you require user and network information, you should use the
        ``fetch_access_data`` method instead.
        """
        access_data = self.fetch_access_data(code)
        try:
            return access_data.access_token.token
        except AttributeError:
            raise ResponseError("Unexpected response format")



# Copyright (c) Microsoft Corporation
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY
# IMPLIED WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR
# PURPOSE, MERCHANTABLITY OR NON-INFRINGEMENT.
#
# See the Apache Version 2.0 License for specific language governing
# permissions and limitations under the License.

from .apis import MessagesAPI, TopicsAPI, UsersAPI, GroupsAPI
from .client import Client


class Yammer(object):

    """
    Main entry point for accessing the Yammer API.

    Essentially this is just a Factory class that provides instances of various
    classes that interact directly with the API. For example, the ``messages``
    method returns a ``MessagesAPI`` object.
    """

    def __init__(self, access_token=None, base_url=None):
        """
        Initialize a new Yammer instance.

        * ``access_token`` identifies the current user. You can acquire an
          access token using an :class:`yampy.Authenticator`.
        * ``base_url`` defaults to the live Yammer API. Provide a different
          base URL to make requests against some other server, e.g. a fake
          in your application's test suite.
        """
        self._client = Client(access_token=access_token, base_url=base_url)

    @property
    def client(self):
        """
        Returns a :class:`yampy.client.Client` object which can be used to make
        HTTP requests to any of the Yammer REST API endpoints.

        You should use this if there isn't a more specific interface available
        for the request you want to make, e.g. if you want to request users you
        should use the ``users`` method instead of the ``client`` method.
        """
        return self._client

    @property
    def messages(self):
        """
        Returns a :class:`yampy.apis.MessagesAPI` object which can be used to
        call the Yammer API's message-related endpoints.
        """
        if not hasattr(self, "_messages_api"):
            self._messages_api = MessagesAPI(client=self._client)
        return self._messages_api

    @property
    def topics(self):
        """
        Returns a :class:`yampy.apis.TopicsAPI` object which can be used to
        call the Yammer API's topic-related endpoints.
        """
        if not hasattr(self, "_topics_api"):
            self._topics_api = TopicsAPI(client=self._client)
        return self._topics_api

    @property
    def users(self):
        """
        Returns a :class:`yampy.apis.UsersAPI` object which can be used to call
        the Yammer API's user-related endpoints.
        """
        if not hasattr(self, "_users_api"):
            self._users_api = UsersAPI(client=self._client)
        return self._users_api

    @property
    def groups(self):
        """
        Returns a :class:`yampy.apis.GroupsAPI` object which can be used to call
        the Yammer API's user-related endpoints.
        """
        if not hasattr(self, "_groups_api"):
            self._groups_api = GroupsAPI(client=self._client)
        return self._groups_api
